/* GENLOCK v.02 */
                  /*******************************/
                  /* GENLOCK */   /* Version .02 */
                  /* Oct. 6, 1990               */
                  /* by Robert Dickow            */
                  /*    1102 E. Sixth St.        */
                  /*    Moscow, ID  83843        */
                  /*    (208)-882-9117           */
                  /*    dickow@idui1 (BITNET)    */
                  /*******************************/

         /*************************************************/
         /*                                               */
         /*  This program allows ARexx control of         */
         /*  Digital Creations SuperGen genlock video     */
         /*  overlay device. This device has manual       */
         /*  controls and comes with hot key and joystick */
         /*  control interfaces. What I needed was some   */
         /*  way to control the genlock from cli/rexx     */
         /*  and programs such as The Director, which has */
         /*  no 'native' genlock control features at this */
         /*  writing. Hence this program.                 */
         /*                                               */
         /*************************************************/

/*--***---***---***---***---***---***---***---***---***---***---***---***--*/
/*                                                                         */
/* Needed to compile this program: ANSII C compiler (Manx 5.0a used here)  */
/*                                 minrexx.c, minrexx.h (from Radical Eye  */
/*                                 but in public domain), supergen.h       */
/*                                 (comes with SuperGen genlock software), */
/*                                 copinit.h (nested in supergen.h)        */
/* Needed to link this program:    Manx 5.0a, dc.lib (comes w/SuperGen),   */
/*                                 c.lib (standard Manx c library, and     */
/*                                 rexxglue.o (comes w/minrexx.c stuff)    */
/* Needed to execute this program: in libs: supergen.library and           */
/*                                 copinit.library, plus the usual rexx    */
/*                                 library stuff to run Arexx, but if you  */
/*                                 plan to get DOS2.0, you will get these  */
/*                                 things already anyhoo.                  */
/*                                                                         */
/*--***---***---***---***---***---***---***---***---***---***---***---***--*/

#include <stdio.h>     /* ...the usual stuff.             */
#include "minrexx.c"   /* Thanks to Radical Eye Software! */
#include <exec/exec.h>
#include <ctype.h>
#include <limits.h>
#include <DCInterface/supergen.h> /* N.B: nested include of copinit.h here */

/***************************************************************/
/* Control the genlock:                                        */
/***************************************************************/

#define GEN_ON   1L     /* Masks... */
#define GEN_OFF  0L

#define MAXPRIORITY 8
#define MINPRIORITY 0

/* Error flags for sending to cleanup() */
#define COULDNT_OPEN_SGENLIB  1

/*--declarations:----------------------------------------------------------*/
/* AREXX STUFF...                                                          */
/* The following are some declarations needed for the ARexx port:          */

char * PortName = "genlock";    /* Name of Rexx msgport for ADDRESS.       */
long PortBit;                   /* For ARexx signal bit.                   */
char * Extension= ".rexx";        /* For default ARexx file extension        */
char * ExtU = "-E";
char * ExtL = "-e";

/* This is the declarations and list of acceptable commands from ARexx
   (and from the CLI as well, but CLI commands are processed separately.)
   The behavior from CLI is execute-and-exit unless
*/

void RXcmd_quit(void);
void RXcmd_Genlock_ON(void), RXcmd_Genlock_OFF(void);
void RXcmd_Priority(void), RXcmd_CallPriority(void);

struct rexxCommandList cmds[] =  {
  { "quit", (APTR) &RXcmd_quit },
  { "genpri",(APTR) &RXcmd_Priority },
  { "on",(APTR) &RXcmd_Genlock_ON },
  { "off",(APTR) &RXcmd_Genlock_OFF },
  { "callpri",(APTR) &RXcmd_CallPriority },
  { NULL, NULL  }
  };

/* Genlock control stuff...                                         */

APTR SuperGenBase= NULL, CopinitBase;

void SuperGen(long Flags, int BackGroundValue, int BGnoVB,
              int GraphicsValue, int GnoVB);
void SuperGenPriority(int Level, int SetLevel);
long StatusGen(void);

int PriorityLevel(void);
void cleanup(int);

/*-------------------------------------------------------------------*/
/* These are some declarations for the general application:          */

/* extern void Wait(long); */

#define QUIT_REQ 1
#define HELP_REQ 2

int errorcode=0;
int all_done=0; /* flag to tell main loop that a quit request is in. */

char *CLIHelp[] = { "-H", "-?", "?", "HELP", "-HELP", NULL };

int rx_args[4] =           /* what args did we get to the command?   */
  { 0, -1, 0, -1 };
int parsed ;               /* # of arguments passed via arexx.       */
void parseargs(char *p, int n);
char returnstr[3]= { 0, 0, 0 };
char *returnbuffer = NULL;
long CallPriority=(long) MAXPRIORITY;
/*-------------------------------------------------------------------*/

/* main() initializes the RexxPort via minrexx.c routines, then waits
          in a loop for more commands until QUIT is issued. If the
          program is executed from the CLI, the REXX polling loop is
          entered only if the CLI command line has the HOLD option flag
          after any CLI option, or if the program is called with no
          command line arguments at all or just HOLD.
          See the READ.ME file for additional CLI and rexx addressable
          commands options.
*/

main(int ncarg, char ** carg)
{
  void RXcmd_handler(struct RexxMsg *, struct rexxCommandList *,char *);
  void init_SuperGen(void);
  int  check_state(char *, char **,int ncrg);

  if (check_state( * (carg+1), carg, ncarg) == FALSE) {
    init_SuperGen();

    PortBit = upRexxPort(PortName,cmds,Extension,&RXcmd_handler);
    while (! all_done) {
      Wait (PortBit);
      dispRexxPort();
    }
    cleanup(errorcode);
  }
}


/* If the user initiates this driver ("genlock") with no argument but
   the program is already resident, then nothing will happen.
   If the user types the arguments "-h", "-?", "help", or "?" then
   a brief help message will be printed and the program will exit. ALL
   other arguments are processed, if recognized as genlock commands, or
   passed to rexx for further processing.
*/

int check_state(char * typed, char ** carg,int ncrg)
{
  char namebuff[81];
  int parseCLIRequest(char * typed, char **,int ncrg);
  int returnval;

  returnval = parseCLIRequest(typed, carg, ncrg);
  if (returnval == HELP_REQ) {
    printf ("\nFrom CLI at any time: genlock [help] | [quit]\n");
    printf ("From CLI at initialization: genlock [help] | [-e .extstr]\n");
    printf ("From rexx: rx genlock[.ext] ON BackGroundValue BGnoVB GraphicsValue GnoVB\n");
    printf ("                            OFF BackGroundValue BGnoVB GraphicsValue GnoVB\n");
    printf ("                            GENPRI <0 - 8>\n");
    printf ("                            CALLPRI <0 - 8>\n");
    printf ("                            QUIT\n");
    returnval = TRUE;
  } else if (returnval == QUIT_REQ) {
      if (FindPort(PortName) != NULL) {
        strcpy (namebuff,"genlock");
        strcat (namebuff,Extension);
        execlp ("rx"," ",namebuff," ","quit"," ",NULL);
        returnval = TRUE;
        }
    } else if (FindPort(PortName) != NULL) {
         returnval = TRUE;
      } else
        returnval = FALSE;
  return (returnval);
}

/* This just sees if cmd line arg is a valid help or quit request... */

int parseCLIRequest(char * typed, char ** carg,int ncrg)
{
  char buff[81];
  int n, index, ishelp, x;
  char *str;

  ishelp = 0;
  if (strlen (typed) <= 80) {
    (void) strcpy(buff,typed);
    n = strlen (buff);
    for (x=0;x<n;x++)
     *(buff+x) = toupper(*(buff+x));
    index = 0;
    while (str = *(CLIHelp+index++)) {
      if (strcmp (buff,str)==0) {
        ishelp = HELP_REQ;
        break;
      }
    }
    if (strcmp(buff,"QUIT")==0) {
      ishelp = QUIT_REQ;
    } else {
      index = 1;
      while (index++ < ncrg) {
        if (strcmp(ExtL,*(carg+index))==0 || strcmp(ExtU,*(carg+index))==0) {
          if (FindPort(PortName) == NULL)
            Extension = *(carg+(++index));
        }
      }
    }
  }
  return (ishelp);
}

/* If Rexx sends a command to this program via the AREXX Port, this routine
   is vectored to handle those commands.
*/
void RXcmd_handler(struct RexxMsg *msg, struct rexxCommandList *dat, char *p)
{
  int x;

  returnbuffer = NULL;
  parseargs((char *)msg->rm_Args[0],4);
  ((int (*)())(dat->userdata))() ;
  replyRexxCmd(msg, 0L, 0L, returnbuffer);
}


/* rexx command routine to set a flag to signal main() loop that
   an exit from the loop has been requested.
*/
void RXcmd_quit(void)
{
  all_done = 1;
}

void init_SuperGen(void)
{

  if ((SuperGenBase=(APTR) OpenLibrary("supergen.library",0L)) == NULL)
    cleanup(COULDNT_OPEN_SGENLIB);
  SuperGenPriority(MAXPRIORITY,TRUE);
}

void cleanup(int errorcode)
{
  dnRexxPort();
  if (SuperGenBase != (long) NULL) {
    RXcmd_Genlock_OFF();
    CloseLibrary(SuperGenBase);
  }
  exit(errorcode);
}

/* Clip passed value to n<9>=0 */
int clip(int val)
{
    if (val > 8)
      val = 8;
    else if (val < 0)
      val = 0;
    return val;
}

int PriorityLevel(void)
{
  int value;
  int count = 0;

  value = ((int)((StatusGen() >>24 ) & 0xff)) ;
  while (value != 0)
    value >>= 1, count++;
  return (count);
}

void Gen_Priority (void)
{
  sprintf ((char *)&returnstr, "%1d",PriorityLevel());
  returnbuffer = (char * )&returnstr;
}

void Gnlck(long Flags)
{
  Flags =  Flags | (CallPriority << 16) ;
  SuperGen(Flags,rx_args[0],rx_args[1],rx_args[2],rx_args[3]);
}

void RXcmd_CallPriority(void)
{
  int val;

  if (parsed >0) {
    val = clip(rx_args[0]);
    CallPriority = val;
  } else {
    sprintf ((char *)&returnstr, "%1d",CallPriority);
    returnbuffer = (char * )&returnstr;
  }
}

void RXcmd_Genlock_ON(void)
{
  Gnlck (GEN_ON);
}

void RXcmd_Genlock_OFF(void)
{
  Gnlck (GEN_OFF);
}

void RXcmd_Priority(void)
{
  int val;

  if (parsed >0) {
    val = rx_args[0];
    val = clip(val);
    SuperGenPriority(PriorityLevel(),FALSE);  /* clear it! */
    SuperGenPriority( val,TRUE);              /* set new priority level */
  } else if (parsed == 0)
      Gen_Priority();
}

/*
 *   This function tries to find `n' signed numeric arguments in the command
 *   string, and stuffs them into the rx_args array.
 */
void parseargs(char *p, int n)
{
   register int i ;
   int x=0;
   char *first_tok;
   char *tok;
   char * strtok(char *,char *);

   while (*p > ' ' && *p)
      p++ ;
   first_tok = p;
   while (((tok=strtok(first_tok," ")) != NULL) && (x < n)) {
     rx_args[x] = atoi(tok);
     first_tok = NULL;
     x++;
   }
   parsed = x;
}

